package uci.inf221.lunarLander.dataStore.component;

import org.cynetec.lib.system.net.ntp.NTPTime;

import uci.inf221.lunarLander.arch.c2.C2Component;
import uci.inf221.lunarLander.arch.c2.C2Notification;
import uci.inf221.lunarLander.arch.c2.C2Request;
import uci.inf221.lunarLander.dataStore.message.ChangeNTPServerRequest;
import uci.inf221.lunarLander.dataStore.message.ClockTickNotification;
import uci.inf221.lunarLander.dataStore.message.CurrentGameStartTimeNotification;
import uci.inf221.lunarLander.dataStore.message.GameStartTimeUpdatedNotification;
import uci.inf221.lunarLander.dataStore.message.GetCurrentDataRequest;
import uci.inf221.lunarLander.dataStore.message.StartGameRequest;
import uci.inf221.lunarLander.dataStore.message.UpdateGameStartTimeRequest;
import uci.inf221.lunarLander.utils.TerminateableLoopThread;

public class ClockComponent extends C2Component {

	public static final String PORT_IDENTIFIER_DATA_STORE_CONNECTOR = "data_store";

	// TODO adjust constants ?

	private static final long NTP_SYNC_INTERVAL = 300000; // 5 minutes

	public static final int TICK_SLEEP_TIME = 2;
	public static final int TICK_INTERVAL = 30;

	private long startTime;
	private NTPTime ntpTime;
	private TickThread tickThread;

	@Override
	protected void createPorts() {
		addBottomPort(PORT_IDENTIFIER_DATA_STORE_CONNECTOR);
	}

	@Override
	protected void initialize() {
		synchronized (this) {
			initNtpTime();
		}

		tickThread = new TickThread();
	}

	@Override
	public void start() {
	}

	private void initNtpTime() {
		ntpTime = new NTPTime();
		ntpTime.setSyncInterval(NTP_SYNC_INTERVAL);
	}

	@Override
	public void stop() {
		if (tickThread.isAlive())
			tickThread.terminate();
		if (ntpTime.isAlive())
			ntpTime.terminate();
	}

	@Override
	public void waitForStop() {
		tickThread.waitForTermination();
		while (ntpTime.isAlive()) {
			try {
				Thread.sleep(1);
			} catch (InterruptedException e) {
			}
		}
	}

	@Override
	protected void receiveNotification(C2Notification notification) {
	}

	@Override
	protected void receiveRequest(C2Request request) {
		if (request instanceof ChangeNTPServerRequest) {
			synchronized (this) {
				if (ntpTime.isAlive()) {
					ntpTime.terminate();
					initNtpTime();
				}

				String server = ((ChangeNTPServerRequest) request)
						.getNtpServer();
				int port = ((ChangeNTPServerRequest) request).getPort();

				ntpTime.setNTPServer(server);
				ntpTime.setNTPServerPort(port);
				ntpTime.start();
			}
		} else if (request instanceof UpdateGameStartTimeRequest) {
			GameStartTimeUpdatedNotification notif = null;
			synchronized (this) {
				if (tickThread.isAlive()) {
					tickThread.terminate();
					tickThread.waitForTermination();
					tickThread = new TickThread();
				}
				startTime = ((UpdateGameStartTimeRequest) request)
						.getStartTime();
				if (startTime != 0) {
					tickThread.start();
					notif = new GameStartTimeUpdatedNotification(this,
							startTime);
				}
			}
			if (notif != null) {
				notif.setInitiatorRequest(request);
				sendNotification(notif);
			}
		} else if (request instanceof GetCurrentDataRequest) {
			CurrentGameStartTimeNotification notif;
			synchronized (this) {
				notif = new CurrentGameStartTimeNotification(this, startTime);
			}
			notif.setInitiatorRequest(request);
			sendNotification(notif);
		} else if (request instanceof StartGameRequest) {
			GameStartTimeUpdatedNotification notif = null;
			synchronized (this) {
				if (tickThread.isAlive()) {
					tickThread.terminate();
					tickThread.waitForTermination();
					tickThread = new TickThread();
				}
				startTime = ntpTime.getNTPTimeMillis();
				tickThread.start();
				notif = new GameStartTimeUpdatedNotification(this, startTime);
			}
			notif.setInitiatorRequest(request);
			sendNotification(notif);
		}
	}

	public class TickThread extends TerminateableLoopThread {
		public TickThread() {
			super(new Runnable() {
				private long lastTickTime;

				@Override
				public void run() {
					long tickTime;
					synchronized (ClockComponent.this) {
						tickTime = (ntpTime.getNTPTimeMillis() - startTime);
					}
					if (tickTime - lastTickTime > TICK_INTERVAL) {
						sendNotification(new ClockTickNotification(
								ClockComponent.this, tickTime));
						lastTickTime = ((long) (tickTime / TICK_INTERVAL))
								* TICK_INTERVAL;
					}
				}
			}, TICK_SLEEP_TIME);
		}
	}

}
