package uci.inf221.lunarLander.gui.component;

import java.awt.Dimension;
import java.awt.KeyEventDispatcher;
import java.awt.KeyboardFocusManager;
import java.awt.Label;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicBoolean;

import javax.swing.JFrame;
import javax.swing.JPanel;

import org.apache.log4j.Logger;

import uci.inf221.lunarLander.arch.c2.C2Component;
import uci.inf221.lunarLander.arch.c2.C2Notification;
import uci.inf221.lunarLander.arch.c2.C2Request;
import uci.inf221.lunarLander.client.Client;
import uci.inf221.lunarLander.dataStore.entities.Asteroid;
import uci.inf221.lunarLander.dataStore.entities.Bullet;
import uci.inf221.lunarLander.dataStore.entities.Level.LevelStatus;
import uci.inf221.lunarLander.dataStore.entities.MoveableDataEntity;
import uci.inf221.lunarLander.dataStore.entities.MovementVector;
import uci.inf221.lunarLander.dataStore.entities.Ship;
import uci.inf221.lunarLander.dataStore.message.ClockTickNotification;
import uci.inf221.lunarLander.dataStore.message.CreateNewBulletRequest;
import uci.inf221.lunarLander.dataStore.message.EntityDestroyedNotification;
import uci.inf221.lunarLander.dataStore.message.LevelStatusUpdatedNotification;
import uci.inf221.lunarLander.dataStore.message.LevelUpdatedNotification;
import uci.inf221.lunarLander.dataStore.message.MovementVectorUpdatedNotification;
import uci.inf221.lunarLander.dataStore.message.NewAsteroidNotification;
import uci.inf221.lunarLander.dataStore.message.NewBulletNotification;
import uci.inf221.lunarLander.dataStore.message.NewShipNotification;
import uci.inf221.lunarLander.dataStore.message.StartGameRequest;
import uci.inf221.lunarLander.dataStore.message.UpdateMovementVectorAccelerationRequest;
import uci.inf221.lunarLander.gui.impl.LevelCanvas;
import uci.inf221.lunarLander.utils.TerminateableLoopThread;

public class GUIComponent extends C2Component {

	public static final String PORT_IDENTIFIER_DATA_STORE_CONNECTOR = "data_store";

	private static final String GAME_TITLE = "Lunar Lander";

	// half a turn per second
	private static final double TURN_SPEED = Math.PI / 1000;
	// px/s² per second
	private static final float ACCELERATION_SPEED = 0.02f / 1000 / 1000;

	private final Logger log = Logger.getLogger(this.getClass());

	private JFrame container;
	private JPanel panel;
	private LevelCanvas levelCanvas;

	private LevelCanvasDrawerThread levelCanvasDrawerThread;

	private Map<UUID, MoveableDataEntity> moveableEntities = new ConcurrentHashMap<UUID, MoveableDataEntity>();
	private Map<UUID, Asteroid> asteroids = new ConcurrentHashMap<UUID, Asteroid>();
	private Map<UUID, Bullet> bullets = new ConcurrentHashMap<UUID, Bullet>();
	private Map<UUID, Ship> ships = new ConcurrentHashMap<UUID, Ship>();

	private LevelStatus levelStatus;
	private Ship playerShip;
	private float[] newPlayerShipAcceleration;

	private Runnable spaceKeyAction;

	private AtomicBoolean isFinished;

	@Override
	protected void createPorts() {
		addTopPort(PORT_IDENTIFIER_DATA_STORE_CONNECTOR);
	}

	@Override
	protected void initialize() {
		isFinished = new AtomicBoolean(false);

		container = new JFrame(GAME_TITLE);
		container.addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing(WindowEvent e) {
				Client.terminate();
			}
		});

		panel = (JPanel) container.getContentPane();
		panel.setPreferredSize(new Dimension(800, 600));
		// panel.setLayout(null);

		levelCanvas = new LevelCanvas(this);

		enterWaitingState();

		container.pack();
		container.setResizable(false);
		container.setLocationRelativeTo(null);

		KeyboardFocusManager.getCurrentKeyboardFocusManager()
				.addKeyEventDispatcher(new GUIKeyListener());
	}

	private synchronized void enterWaitingState() {
		isFinished.set(false);

		clearPanel();
		panel.add(new Label("Waiting for Connection ...", Label.CENTER));
		panel.revalidate();
	}

	private synchronized void enterReadyState() {
		isFinished.set(false);

		clearPanel();
		panel.add(new Label("Ready. Hit space to start!", Label.CENTER));
		panel.revalidate();

		spaceKeyAction = new Runnable() {
			@Override
			public void run() {
				sendRequest(new StartGameRequest(GUIComponent.this));
			}
		};
	}

	private synchronized void enterStartedState() {
		isFinished.set(false);

		clearPanel();
		panel.add(levelCanvas);
		levelCanvas.createBufferStrategy(2);
		panel.revalidate();

		spaceKeyAction = new Runnable() {
			@Override
			public void run() {
				if (playerShip != null) {
					Bullet bullet = new Bullet();
					MovementVector movementVector = new MovementVector();
					float[] position = playerShip.getMovementVector()
							.getPosition().clone();
					float dir = playerShip.getMovementVector()
							.getAccelerationDirection();
					float mag = Ship.EXTENT + 1;
					position[0] += mag * Math.cos(dir);
					position[1] += mag * Math.sin(dir);
					movementVector.setPosition(position);
					movementVector.setSpeed(dir, Bullet.SPEED_MAGNITUTE);
					movementVector.setAcceleration(0, 0);
					bullet.setMovementVector(movementVector);
					bullet.setOriginEntityId(playerShip.getId());
					sendRequest(new CreateNewBulletRequest(GUIComponent.this,
							bullet));
				}
			}
		};

		levelCanvasDrawerThread = new LevelCanvasDrawerThread();
		levelCanvasDrawerThread.start();
	}

	private synchronized void enterResultState(LevelStatus levelStatus) {
		isFinished.set(true);
		newPlayerShipAcceleration = null;

		clearPanel();
		moveableEntities.clear();
		asteroids.clear();
		bullets.clear();
		ships.clear();

		Runnable continueRunnable = new Runnable() {
			private AtomicBoolean running = new AtomicBoolean(false);

			@Override
			public void run() {
				if (!running.getAndSet(true)) {
					LevelStatus levelStatus;
					synchronized (GUIComponent.this) {
						levelStatus = GUIComponent.this.levelStatus;
					}

					for (int i = 0; i < 5; i++) {
						if (levelStatus.hasEnded()) {
							try {
								// sleep a bit so that new level status can be
								// received
								Thread.sleep(100);
							} catch (InterruptedException e) {
							}

							synchronized (GUIComponent.this) {
								levelStatus = GUIComponent.this.levelStatus;
							}
						} else
							break;
					}

					if (levelStatus == LevelStatus.LOADED
							|| levelStatus == LevelStatus.INITIALISING) {
						enterWaitingState();
					} else if (levelStatus == LevelStatus.READY) {
						enterReadyState();
					} else {
						// no more levels
						Client.terminate();
					}
				}
			}
		};

		switch (levelStatus) {
		case COMPLETED:
			panel.add(new Label("Congratulations! Hit space to continue.",
					Label.CENTER));

			spaceKeyAction = continueRunnable;
			break;
		case ERROR:
			panel.add(new Label("Connection lost. Retrying ...", Label.CENTER));
			break;
		case FAILED:
			panel.add(new Label("Aww, too bad. Hit space to continue.",
					Label.CENTER));

			spaceKeyAction = continueRunnable;
			break;
		default:
			break;
		}

		panel.revalidate();
	}

	private void clearPanel() {
		if (levelCanvasDrawerThread != null
				&& levelCanvasDrawerThread.isAlive()) {
			levelCanvasDrawerThread.terminate();
			levelCanvasDrawerThread.waitForTermination();
		}
		panel.removeAll();
		spaceKeyAction = null;
	}

	@Override
	public void start() {
		container.setVisible(true);
	}

	@Override
	public synchronized void stop() {
		if (levelCanvasDrawerThread != null
				&& levelCanvasDrawerThread.isAlive()) {
			levelCanvasDrawerThread.terminate();
		}
		container.setVisible(false);
		container.dispose();
	}

	@Override
	public synchronized void waitForStop() {
		if (levelCanvasDrawerThread != null
				&& levelCanvasDrawerThread.isAlive()) {
			levelCanvasDrawerThread.waitForTermination();
		}
	}

	@Override
	protected void receiveNotification(C2Notification notification) {
		if (notification instanceof LevelStatusUpdatedNotification) {
			LevelStatus levelStatus = ((LevelStatusUpdatedNotification) notification)
					.getLevelStatus();
			synchronized (this) {
				this.levelStatus = levelStatus;
			}
			if (levelStatus.hasEnded()) {
				enterResultState(levelStatus);
			} else if (levelStatus == LevelStatus.READY) {
				if (!isFinished.get()) {
					enterReadyState();
				}
			} else if (levelStatus == LevelStatus.STARTED) {
				enterStartedState();
			}
		} else if (notification instanceof LevelUpdatedNotification) {
			synchronized (this) {
				levelCanvas.setLevel(((LevelUpdatedNotification) notification)
						.getLevel());
			}
		} else if (notification instanceof MovementVectorUpdatedNotification) {
			synchronized (this) {
				MoveableDataEntity entity = moveableEntities
						.get(((MovementVectorUpdatedNotification) notification)
								.getEntityId());
				if (entity != null) {
					MovementVector movementVector = ((MovementVectorUpdatedNotification) notification)
							.getMovementVector();
					if (entity == playerShip
							&& newPlayerShipAcceleration != null) {
						movementVector
								.setAcceleration(newPlayerShipAcceleration);
					}
					entity.setMovementVector(movementVector);
				} else {
					log.warn("movement vector update for unkown entity of id: "
							+ ((MovementVectorUpdatedNotification) notification)
									.getEntityId());
				}
			}
		} else if (notification instanceof NewAsteroidNotification) {
			synchronized (this) {
				Asteroid asteroid = ((NewAsteroidNotification) notification)
						.getAsteroid();
				moveableEntities.put(asteroid.getId(), asteroid);
				asteroids.put(asteroid.getId(), asteroid);
			}
		} else if (notification instanceof NewBulletNotification) {
			synchronized (this) {
				Bullet bullet = ((NewBulletNotification) notification)
						.getBullet();
				moveableEntities.put(bullet.getId(), bullet);
				bullets.put(bullet.getId(), bullet);
			}
		} else if (notification instanceof NewShipNotification) {
			synchronized (this) {
				Ship ship = ((NewShipNotification) notification).getShip();
				moveableEntities.put(ship.getId(), ship);
				ships.put(ship.getId(), ship);
				if (ship.getPlayerId().equals(Client.getPlayerId())) {
					playerShip = ship;
				}
			}
		} else if (notification instanceof EntityDestroyedNotification) {
			synchronized (this) {
				moveableEntities
						.remove(((EntityDestroyedNotification) notification)
								.getEntityId());
				ships.remove(((EntityDestroyedNotification) notification)
						.getEntityId());
				bullets.remove(((EntityDestroyedNotification) notification)
						.getEntityId());
				asteroids.remove(((EntityDestroyedNotification) notification)
						.getEntityId());

				if (playerShip != null
						&& playerShip.getId().equals(
								((EntityDestroyedNotification) notification)
										.getEntityId())) {
					playerShip = null;
					newPlayerShipAcceleration = null;
				}

				// TODO destroy animation
			}
		} else if (notification instanceof ClockTickNotification) {
			synchronized (this) {
				if (newPlayerShipAcceleration != null) {
					if (playerShip != null) {
						log.trace("sending new player ship acceleration: ("
								+ newPlayerShipAcceleration[0] + ", "
								+ newPlayerShipAcceleration[1] + ")");
						sendRequest(new UpdateMovementVectorAccelerationRequest(
								this, playerShip.getId(),
								newPlayerShipAcceleration));
					}
					newPlayerShipAcceleration = null;
				}
			}
		}
	}

	@Override
	protected void receiveRequest(C2Request request) {
	}

	private class LevelCanvasDrawerThread extends TerminateableLoopThread {

		private static final int SLEEP_TIME = 10;

		public LevelCanvasDrawerThread() {
			super(new Runnable() {
				@Override
				public void run() {
					Ship[] shipArray;
					Asteroid[] asteroidArray;
					Bullet[] bulletArray;
					synchronized (GUIComponent.this) {
						shipArray = ships.values().toArray(
								new Ship[ships.size()]);
						asteroidArray = asteroids.values().toArray(
								new Asteroid[asteroids.size()]);
						bulletArray = bullets.values().toArray(
								new Bullet[bullets.size()]);
					}

					levelCanvas.drawContents(shipArray, asteroidArray,
							bulletArray);
				}
			}, SLEEP_TIME);
		}

	}

	private class GUIKeyListener implements KeyEventDispatcher {

		private AtomicBoolean releasedSpace = new AtomicBoolean(true);

		@Override
		public boolean dispatchKeyEvent(KeyEvent e) {
			if (e.getID() == KeyEvent.KEY_PRESSED) {
				if (e.getKeyCode() == KeyEvent.VK_ESCAPE) {
					Client.terminate();
					return true;
				} else if (e.getKeyCode() == KeyEvent.VK_SPACE) {
					if (releasedSpace.getAndSet(false)) {
						synchronized (GUIComponent.this) {
							if (spaceKeyAction != null)
								spaceKeyAction.run();
						}
					}
				} else if (e.getKeyCode() == KeyEvent.VK_LEFT) {
					levelCanvas.getLeftKeyPressed().set(true);
				} else if (e.getKeyCode() == KeyEvent.VK_RIGHT) {
					levelCanvas.getRightKeyPressed().set(true);
				} else if (e.getKeyCode() == KeyEvent.VK_UP) {
					levelCanvas.getUpKeyPressed().set(true);
				} else if (e.getKeyCode() == KeyEvent.VK_DOWN) {
					levelCanvas.getDownKeyPressed().set(true);
				}
			} else if (e.getID() == KeyEvent.KEY_RELEASED) {
				if (e.getKeyCode() == KeyEvent.VK_SPACE) {
					releasedSpace.set(true);
				} else if (e.getKeyCode() == KeyEvent.VK_LEFT) {
					levelCanvas.getLeftKeyPressed().set(false);
				} else if (e.getKeyCode() == KeyEvent.VK_RIGHT) {
					levelCanvas.getRightKeyPressed().set(false);
				} else if (e.getKeyCode() == KeyEvent.VK_UP) {
					levelCanvas.getUpKeyPressed().set(false);
				} else if (e.getKeyCode() == KeyEvent.VK_DOWN) {
					levelCanvas.getDownKeyPressed().set(false);
				}
			}
			return false;
		}

	}

	public void turnPlayerShipLeft(long timeDiff) {
		float turnAngle = (float) (timeDiff * TURN_SPEED);
		synchronized (this) {
			if (playerShip != null) {
				MovementVector movementVector = playerShip.getMovementVector();
				movementVector.turnAcceleration(-turnAngle);
				newPlayerShipAcceleration = movementVector.getAcceleration();
				log.trace("new player ship acceleration: ("
						+ newPlayerShipAcceleration[0] + ", "
						+ newPlayerShipAcceleration[1] + ")");
			}
		}
	}

	public void turnPlayerShipRight(long timeDiff) {
		turnPlayerShipLeft(-timeDiff);
	}

	public void acceleratePlayerShip(long timeDiff) {
		float incAcceleration = timeDiff * ACCELERATION_SPEED;
		synchronized (this) {
			if (playerShip != null) {
				MovementVector movementVector = playerShip.getMovementVector();
				movementVector.addAcceleration(incAcceleration);
				if (movementVector.getAccelerationMagnitude() > Ship.MAX_ACCELERATION) {
					movementVector.setAcceleration(
							movementVector.getAccelerationDirection(),
							Ship.MAX_ACCELERATION);
				} else if (movementVector.getAccelerationMagnitude() < 0) {
					movementVector.setAcceleration(
							movementVector.getAccelerationDirection(), 0);
				}
				newPlayerShipAcceleration = movementVector.getAcceleration();
				log.trace("new player ship acceleration: ("
						+ newPlayerShipAcceleration[0] + ", "
						+ newPlayerShipAcceleration[1] + ")");
			}
		}
	}

	public void deceleratePlayerShip(long timeDiff) {
		acceleratePlayerShip(-timeDiff);
	}

	public void sizeUpdated(Dimension size) {
		if (container.isVisible()) {
			panel.setPreferredSize(size);
			container.pack();
		}
	}

}
