package uci.inf221.lunarLander.gui.impl;

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GraphicsConfiguration;
import java.awt.GraphicsEnvironment;
import java.awt.Transparency;
import java.awt.image.BufferStrategy;
import java.awt.image.BufferedImage;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicReference;

import uci.inf221.lunarLander.client.Client;
import uci.inf221.lunarLander.dataStore.entities.Asteroid;
import uci.inf221.lunarLander.dataStore.entities.Bullet;
import uci.inf221.lunarLander.dataStore.entities.Level;
import uci.inf221.lunarLander.dataStore.entities.MovementVector;
import uci.inf221.lunarLander.dataStore.entities.Ship;
import uci.inf221.lunarLander.gui.component.GUIComponent;

public class LevelCanvas extends Canvas {

	private class LevelInfo {
		private BufferedImage image;
		private int width;
		private int height;

		public LevelInfo(BufferedImage image, int width, int height) {
			this.image = image;
			this.width = width;
			this.height = height;
		}

		public BufferedImage getImage() {
			return image;
		}

		public int getWidth() {
			return width;
		}

		public int getHeight() {
			return height;
		}
	}

	private static final Color SKY_COLOR = new Color(50, 50, 50);
	private static final Color MOON_COLOR = new Color(200, 200, 200);
	private static final long serialVersionUID = 5504826605591041489L;

	private GUIComponent guiComponent;

	private AtomicReference<LevelInfo> levelInfo = new AtomicReference<LevelInfo>();
	private AtomicBoolean leftKeyPressed = new AtomicBoolean(false);
	private AtomicBoolean rightKeyPressed = new AtomicBoolean(false);
	private AtomicBoolean upKeyPressed = new AtomicBoolean(false);
	private AtomicBoolean downKeyPressed = new AtomicBoolean(false);

	private long lastDrawTime = 0;

	public LevelCanvas(GUIComponent guiComponent) {
		super();

		this.guiComponent = guiComponent;
		setBounds(0, 0, 800, 600);

		setIgnoreRepaint(true);
	}

	public void setLevel(Level level) {
		GraphicsConfiguration gc = GraphicsEnvironment
				.getLocalGraphicsEnvironment().getDefaultScreenDevice()
				.getDefaultConfiguration();
		int width = level.getWidth();
		int height = level.getHeight();
		BufferedImage image = gc.createCompatibleImage(width, height,
				Transparency.BITMASK);

		Graphics graphics = image.getGraphics();
		graphics.setColor(new Color(0, 0, 0, 0));
		graphics.fillRect(0, 0, width, height);

		graphics.setColor(MOON_COLOR);

		int x = 0;
		int y = 0;
		for (boolean[] column : level.getMap()) {
			for (boolean pixel : column) {
				if (pixel) {
					graphics.fillRect(x, y, 1, 1);
				}
				y++;
			}
			x++;
			y = 0;
		}

		levelInfo.set(new LevelInfo(image, width, height));
	}

	public void drawContents(Ship[] ships, Asteroid[] asteroids,
			Bullet[] bullets) {
		LevelInfo levelInfo = this.levelInfo.get();
		if (levelInfo == null)
			return;

		long timeDiff = System.currentTimeMillis() - lastDrawTime;
		lastDrawTime = lastDrawTime + timeDiff;

		handleInputs(timeDiff);

		if (getWidth() != levelInfo.getWidth()
				|| getHeight() != levelInfo.getHeight()) {
			setBounds(0, 0, levelInfo.getWidth(), levelInfo.getHeight());
			guiComponent.sizeUpdated(this.getSize());
		}

		BufferStrategy strategy = getBufferStrategy();

		Graphics2D g = (Graphics2D) strategy.getDrawGraphics();
		g.setColor(SKY_COLOR);
		g.fillRect(0, 0, 800, 600);

		drawLevel(g, levelInfo);

		drawAsteroids(g, asteroids);
		drawBullets(g, bullets);
		drawShips(g, ships);

		g.dispose();
		strategy.show();
	}

	private void handleInputs(long timeDiff) {
		if (leftKeyPressed.get() && !rightKeyPressed.get()) {
			guiComponent.turnPlayerShipLeft(timeDiff);
		} else if (rightKeyPressed.get() && !leftKeyPressed.get()) {
			guiComponent.turnPlayerShipRight(timeDiff);
		}

		if (upKeyPressed.get() && !downKeyPressed.get()) {
			guiComponent.acceleratePlayerShip(timeDiff);
		} else if (downKeyPressed.get() && !upKeyPressed.get()) {
			guiComponent.deceleratePlayerShip(timeDiff);
		}
	}

	private void drawShips(Graphics2D g, Ship[] ships) {
		Ship playerShip = null;
		for (Ship ship : ships) {
			if (ship.getPlayerId().equals(Client.getPlayerId())) {
				playerShip = ship;
			} else {
				drawShip(g, ship, true);
			}
		}

		// draw playerShip last
		if (playerShip != null) {
			drawShip(g, playerShip, false);
		}
	}

	private void drawShip(Graphics2D g, Ship ship, boolean alpha) {
		MovementVector movementVector = ship.getMovementVector();
		float[] position = movementVector.getPosition();
		Sprite sprite = SpriteStore.getInstance().getSprite(
				SpriteStore.SPRITE_SHIP);
		if (alpha)
			sprite = sprite.getTranslucentSprite();
		sprite.drawRotated(g, movementVector.getAccelerationDirection(),
				position[0], position[1]);

		int thrustSprite = (int) Math.floor(movementVector
				.getAccelerationMagnitude()
				/ Ship.MAX_ACCELERATION
				* SpriteStore.SPRITE_THRUST.length);
		if (thrustSprite >= SpriteStore.SPRITE_THRUST.length)
			thrustSprite = SpriteStore.SPRITE_THRUST.length - 1;
		sprite = SpriteStore.getInstance().getSprite(
				SpriteStore.SPRITE_THRUST[thrustSprite]);
		sprite.drawRotated(g, movementVector.getAccelerationDirection(),
				position[0], position[1], -16, 0);
	}

	private void drawBullets(Graphics2D g, Bullet[] bullets) {
		for (Bullet bullet : bullets) {
			float[] position = bullet.getMovementVector().getPosition();
			Sprite sprite = SpriteStore.getInstance().getSprite(
					SpriteStore.SPRITE_BULLET);
			sprite.draw(g, position[0], position[1]);
		}
	}

	private void drawAsteroids(Graphics2D g, Asteroid[] asteroids) {
		for (Asteroid asteroid : asteroids) {
			float[] position = asteroid.getMovementVector().getPosition();
			Sprite sprite = SpriteStore.getInstance().getSprite(
					SpriteStore.SPRITE_ASTEROID);
			sprite.draw(g, position[0], position[1]);
		}
	}

	private void drawLevel(Graphics2D g, LevelInfo levelInfo) {
		g.drawImage(levelInfo.getImage(), 0, 0, null);
	}

	public AtomicBoolean getLeftKeyPressed() {
		return leftKeyPressed;
	}

	public AtomicBoolean getRightKeyPressed() {
		return rightKeyPressed;
	}

	public AtomicBoolean getUpKeyPressed() {
		return upKeyPressed;
	}

	public AtomicBoolean getDownKeyPressed() {
		return downKeyPressed;
	}
}
