package uci.inf221.lunarLander.logic.component;

import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicLong;

import org.apache.log4j.Logger;

import uci.inf221.lunarLander.arch.Port.PortDirection;
import uci.inf221.lunarLander.arch.c2.C2Component;
import uci.inf221.lunarLander.arch.c2.C2Notification;
import uci.inf221.lunarLander.arch.c2.C2Request;
import uci.inf221.lunarLander.dataStore.entities.Asteroid;
import uci.inf221.lunarLander.dataStore.entities.Bullet;
import uci.inf221.lunarLander.dataStore.entities.Level;
import uci.inf221.lunarLander.dataStore.entities.Level.LevelStatus;
import uci.inf221.lunarLander.dataStore.entities.MoveableDataEntity;
import uci.inf221.lunarLander.dataStore.entities.Ship;
import uci.inf221.lunarLander.dataStore.message.ClockTickNotification;
import uci.inf221.lunarLander.dataStore.message.EntityDestroyedNotification;
import uci.inf221.lunarLander.dataStore.message.LevelStatusUpdatedNotification;
import uci.inf221.lunarLander.dataStore.message.LevelUpdatedNotification;
import uci.inf221.lunarLander.dataStore.message.MovementVectorUpdatedNotification;
import uci.inf221.lunarLander.dataStore.message.NewAsteroidNotification;
import uci.inf221.lunarLander.dataStore.message.NewBulletNotification;
import uci.inf221.lunarLander.dataStore.message.NewShipNotification;
import uci.inf221.lunarLander.logic.message.EntityCollisionNotification;
import uci.inf221.lunarLander.logic.message.EntityLevelCollisionNotification;
import uci.inf221.lunarLander.logic.message.EntityOutsideMapNotification;
import uci.inf221.lunarLander.logic.message.ShipTouchingNotification;
import uci.inf221.lunarLander.utils.TerminateableLoopThread;

public class CollisionDetectionLogicComponent extends C2Component {

	public static final String PORT_IDENTIFIER_DATA_STORE_CONNECTOR = "data_store";
	public static final String PORT_IDENTIFIER_LOGIC_CONNECTOR = "logic";

	private final Logger log = Logger.getLogger(this.getClass());

	private Map<UUID, MoveableDataEntity> moveableEntities = new ConcurrentHashMap<UUID, MoveableDataEntity>();

	private AtomicLong nextClockTick = new AtomicLong(-1);
	private TerminateableLoopThread collisionDetectionThread;
	private Level level;

	@Override
	protected void createPorts() {
		addTopPort(PORT_IDENTIFIER_DATA_STORE_CONNECTOR);
		addBottomPort(PORT_IDENTIFIER_LOGIC_CONNECTOR).setDirection(
				PortDirection.OUT);
	}

	@Override
	protected void initialize() {
		collisionDetectionThread = new TerminateableLoopThread(new Runnable() {
			@Override
			public void run() {
				while (nextClockTick.getAndSet(-1) != -1) {
					try {
						Thread.sleep(10);
					} catch (InterruptedException e) {
					}
				}
				detectCollisions();
			}
		});
	}

	@Override
	public void start() {
		collisionDetectionThread.start();
	}

	@Override
	public void stop() {
		collisionDetectionThread.terminate();
	}

	@Override
	public void waitForStop() {
		collisionDetectionThread.waitForTermination();
	}

	@SuppressWarnings("unchecked")
	private void detectCollisions() {
		Level level;
		synchronized (this) {
			level = this.level;
			if (level == null)
				return;
		}

		boolean[][] levelMap = level.getMap();
		int width = levelMap.length;
		if (width < 1)
			return;
		int height = levelMap[0].length;

		Object[][] collisionMap = new Object[width][height];

		Set<Set<UUID>> collisions = new HashSet<Set<UUID>>();

		for (Entry<UUID, MoveableDataEntity> entry : moveableEntities
				.entrySet()) {
			UUID entityId = entry.getKey();
			MoveableDataEntity entity = entry.getValue();

			float[] position;
			synchronized (entity) {
				position = entity.getMovementVector().getPosition();
			}
			int extent = entity.getExtent();

			if (position[0] > width - 1 || position[0] < 0
					|| position[1] > height - 1 || position[1] < 0) {
				sendNotification(new EntityOutsideMapNotification(this,
						entityId));
			} else {
				// x = 2 => 0 (incl) to 4 (excl)
				// x = 2.1 => 1 (incl) to 5 (excl)
				int xStart = (int) Math.ceil(position[0] - (extent / 2));
				int xEnd = (int) Math.ceil(position[0] + (extent / 2));
				int yStart = (int) Math.ceil(position[1] - (extent / 2));
				int yEnd = (int) Math.ceil(position[1] + (extent / 2));

				if (xStart < 0)
					xStart = 0;
				if (yStart < 0)
					yStart = 0;
				if (xEnd >= width)
					xEnd = width - 1;
				if (yEnd >= height)
					yEnd = height - 1;

				boolean shipTouching = (entity instanceof Ship);
				boolean levelCollision = false;

				for (int x = xStart; x < xEnd; x++) {
					for (int y = yStart; y < yEnd; y++) {
						if (levelMap[x][y]) {
							shipTouching &= y < yEnd;
							levelCollision = true;
						}

						if (collisionMap[x][y] == null) {
							collisionMap[x][y] = new HashSet<UUID>();
							((HashSet<UUID>) collisionMap[x][y]).add(entityId);
						} else {
							((HashSet<UUID>) collisionMap[x][y]).add(entityId);

							for (UUID otherEntity : ((HashSet<UUID>) collisionMap[x][y])) {
								if (!otherEntity.equals(entityId)) {
									HashSet<UUID> collision = new HashSet<UUID>();
									collision.add(entityId);
									collision.add(otherEntity);
									collisions.add(collision);
								}
							}
						}
					}
				}

				if (levelCollision && shipTouching) {
					sendNotification(new ShipTouchingNotification(this,
							entityId));
				} else if (levelCollision) {
					sendNotification(new EntityLevelCollisionNotification(this,
							entityId));
				}
			}
		}

		for (Set<UUID> collision : collisions) {
			UUID[] entityIds = collision.toArray(new UUID[2]);
			sendNotification(new EntityCollisionNotification(this,
					entityIds[0], entityIds[1]));
		}
	}

	@Override
	protected void receiveNotification(C2Notification notification) {
		if (notification instanceof ClockTickNotification) {
			nextClockTick.set(((ClockTickNotification) notification)
					.getTickTime());
		} else if (notification instanceof MovementVectorUpdatedNotification) {
			MoveableDataEntity entity = moveableEntities
					.get(((MovementVectorUpdatedNotification) notification)
							.getEntityId());
			if (entity != null) {
				synchronized (entity) {
					entity.setMovementVector(((MovementVectorUpdatedNotification) notification)
							.getMovementVector());
				}
			} else {
				log.warn("movement vector update for unkown entity of id: "
						+ ((MovementVectorUpdatedNotification) notification)
								.getEntityId());
			}
		} else if (notification instanceof NewAsteroidNotification) {
			Asteroid asteroid = ((NewAsteroidNotification) notification)
					.getAsteroid();
			moveableEntities.put(asteroid.getId(), asteroid);
		} else if (notification instanceof NewBulletNotification) {
			Bullet bullet = ((NewBulletNotification) notification).getBullet();
			moveableEntities.put(bullet.getId(), bullet);
		} else if (notification instanceof NewShipNotification) {
			Ship ship = ((NewShipNotification) notification).getShip();
			moveableEntities.put(ship.getId(), ship);
		} else if (notification instanceof EntityDestroyedNotification) {
			moveableEntities
					.remove(((EntityDestroyedNotification) notification)
							.getEntityId());
		} else if (notification instanceof LevelStatusUpdatedNotification) {
			LevelStatus levelStatus = ((LevelStatusUpdatedNotification) notification)
					.getLevelStatus();
			if (levelStatus.hasEnded()) {
				moveableEntities.clear();
			}
		} else if (notification instanceof LevelUpdatedNotification) {
			synchronized (this) {
				level = ((LevelUpdatedNotification) notification).getLevel();
			}
		}
	}

	@Override
	protected void receiveRequest(C2Request request) {
	}

}
