package uci.inf221.lunarLander.logic.component;

import java.util.Map;
import java.util.Map.Entry;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicLong;

import org.apache.log4j.Logger;

import uci.inf221.lunarLander.arch.c2.C2Component;
import uci.inf221.lunarLander.arch.c2.C2Notification;
import uci.inf221.lunarLander.arch.c2.C2Request;
import uci.inf221.lunarLander.dataStore.entities.Asteroid;
import uci.inf221.lunarLander.dataStore.entities.Bullet;
import uci.inf221.lunarLander.dataStore.entities.Level.LevelStatus;
import uci.inf221.lunarLander.dataStore.entities.MovementVector;
import uci.inf221.lunarLander.dataStore.entities.Ship;
import uci.inf221.lunarLander.dataStore.message.ClockTickNotification;
import uci.inf221.lunarLander.dataStore.message.EntityDestroyedNotification;
import uci.inf221.lunarLander.dataStore.message.LevelStatusUpdatedNotification;
import uci.inf221.lunarLander.dataStore.message.MovementVectorUpdatedNotification;
import uci.inf221.lunarLander.dataStore.message.NewAsteroidNotification;
import uci.inf221.lunarLander.dataStore.message.NewBulletNotification;
import uci.inf221.lunarLander.dataStore.message.NewShipNotification;
import uci.inf221.lunarLander.dataStore.message.UpdateMovementVectorPositionAndSpeedRequest;
import uci.inf221.lunarLander.utils.TerminateableLoopThread;

public class PositionLogicComponent extends C2Component {

	public static final String PORT_IDENTIFIER_DATA_STORE_CONNECTOR = "data_store";

	private Logger log = Logger.getLogger(this.getClass());

	private Map<UUID, MovementVector> movementVectors = new ConcurrentHashMap<UUID, MovementVector>();

	private AtomicLong nextClockTick = new AtomicLong(-1);
	private TerminateableLoopThread updatePositionsThread;

	@Override
	protected void createPorts() {
		addTopPort(PORT_IDENTIFIER_DATA_STORE_CONNECTOR);
	}

	@Override
	protected void initialize() {
		updatePositionsThread = new TerminateableLoopThread(new Runnable() {
			@Override
			public void run() {
				long curClockTick;
				while ((curClockTick = nextClockTick.getAndSet(-1)) == -1
						&& !updatePositionsThread.isTerminated()) {
					try {
						Thread.sleep(2);
					} catch (InterruptedException e) {
					}
				}
				if (curClockTick != -1)
					updatePositions(curClockTick);
			}
		});
	}

	@Override
	public void start() {
		updatePositionsThread.start();
	}

	@Override
	public void stop() {
		updatePositionsThread.terminate();
	}

	@Override
	public void waitForStop() {
		updatePositionsThread.waitForTermination();
	}

	private void updatePositions(long curClockTick) {
		for (Entry<UUID, MovementVector> entry : movementVectors.entrySet()) {
			synchronized (this) {
				MovementVector oldVector = entry.getValue();
				MovementVector newVector = oldVector.moveToTime(curClockTick);
				entry.setValue(newVector);
				sendRequest(new UpdateMovementVectorPositionAndSpeedRequest(
						this, entry.getKey(), newVector.getPosition(),
						newVector.getSpeed(), newVector.getClockTick()));
				float[] oldPos = oldVector.getPosition();
				float[] newPos = oldVector.getPosition();
				log.trace("updated movement vector for entity "
						+ entry.getKey() + " old pos: (" + oldPos[0] + ", "
						+ oldPos[1] + ") new pos: (" + newPos[0] + ", "
						+ newPos[1] + ")");
			}
		}
	}

	@Override
	protected void receiveNotification(C2Notification notification) {
		if (notification instanceof ClockTickNotification) {
			nextClockTick.set(((ClockTickNotification) notification)
					.getTickTime());
		} else if (notification instanceof MovementVectorUpdatedNotification) {
			if (notification.getInitiator() != this) {
				synchronized (this) {
					movementVectors.put(
							((MovementVectorUpdatedNotification) notification)
									.getEntityId(),
							((MovementVectorUpdatedNotification) notification)
									.getMovementVector());
				}
			}
		} else if (notification instanceof NewAsteroidNotification) {
			Asteroid asteroid = ((NewAsteroidNotification) notification)
					.getAsteroid();
			movementVectors.put(asteroid.getId(), asteroid.getMovementVector());
		} else if (notification instanceof NewBulletNotification) {
			Bullet bullet = ((NewBulletNotification) notification).getBullet();
			movementVectors.put(bullet.getId(), bullet.getMovementVector());
		} else if (notification instanceof NewShipNotification) {
			Ship ship = ((NewShipNotification) notification).getShip();
			movementVectors.put(ship.getId(), ship.getMovementVector());
		} else if (notification instanceof EntityDestroyedNotification) {
			movementVectors.remove(((EntityDestroyedNotification) notification)
					.getEntityId());
		} else if (notification instanceof LevelStatusUpdatedNotification) {
			LevelStatus levelStatus = ((LevelStatusUpdatedNotification) notification)
					.getLevelStatus();
			if (levelStatus.hasEnded()) {
				movementVectors.clear();
			}
		}
	}

	@Override
	protected void receiveRequest(C2Request request) {
	}

}
