package uci.inf221.lunarLander.managing.component;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;

import org.apache.commons.io.FileUtils;
import org.apache.log4j.Logger;

import uci.inf221.lunarLander.arch.c2.C2Component;
import uci.inf221.lunarLander.arch.c2.C2Notification;
import uci.inf221.lunarLander.arch.c2.C2Request;
import uci.inf221.lunarLander.dataStore.entities.Level.LevelStatus;
import uci.inf221.lunarLander.dataStore.message.EntityDestroyedNotification;
import uci.inf221.lunarLander.dataStore.message.LevelStatusUpdatedNotification;
import uci.inf221.lunarLander.dataStore.message.NewShipNotification;
import uci.inf221.lunarLander.dataStore.message.UpdateGameStartTimeRequest;
import uci.inf221.lunarLander.dataStore.message.UpdateLevelMapRequest;
import uci.inf221.lunarLander.dataStore.message.UpdateLevelStatusRequest;
import uci.inf221.lunarLander.server.Server;

public class LevelManagerComponent extends C2Component {

	public static final String PORT_IDENTIFIER_DATA_STORE_CONNECTOR = "data_store";
	private final Logger log = Logger.getLogger(this.getClass());

	private Set<UUID> ships = Collections
			.newSetFromMap(new ConcurrentHashMap<UUID, Boolean>());

	private Timer timer = new Timer();
	private AtomicBoolean loaded = new AtomicBoolean(false);
	private AtomicBoolean ended = new AtomicBoolean(false);
	private List<String> levels;
	private AtomicInteger levelNumber = new AtomicInteger(0);

	@Override
	protected void createPorts() {
		addTopPort(PORT_IDENTIFIER_DATA_STORE_CONNECTOR);
	}

	@Override
	protected void initialize() {
		levels = new ArrayList<String>();
		try {
			ArrayList<File> levelFiles = new ArrayList<File>(
					FileUtils.listFiles(new File("levels/"),
							new String[] { "lvl" }, false));

			Collections.sort(levelFiles);

			for (File file : levelFiles) {
				levels.add(FileUtils.readFileToString(file));
			}
		} catch (IOException e) {
			log.error("could not read level file(s)", e);
			throw new RuntimeException("could not read level file(s)", e);
		}
	}

	@Override
	public void start() {
		sendRequest(new UpdateLevelMapRequest(this, levels.get(levelNumber
				.get())));
	}

	@Override
	public void stop() {
		timer.cancel();
	}

	@Override
	public void waitForStop() {
	}

	@Override
	protected void receiveNotification(C2Notification notification) {
		if (notification instanceof NewShipNotification) {
			ships.add(((NewShipNotification) notification).getShip().getId());
			if (ships.size() >= Server.MIN_PLAYERS && loaded.get()) { // TODO
																		// change
																		// back
																		// to >
																		// 1
				sendRequest(new UpdateLevelStatusRequest(this,
						LevelStatus.READY));
			}
		} else if (notification instanceof EntityDestroyedNotification) {
			ships.remove(((EntityDestroyedNotification) notification)
					.getEntityId());
			if (ships.isEmpty() && !ended.get()) {
				sendRequest(new UpdateLevelStatusRequest(this,
						LevelStatus.FAILED));
				sendRequest(new UpdateGameStartTimeRequest(this, 0));
			}
		} else if (notification instanceof LevelStatusUpdatedNotification) {
			LevelStatus status = ((LevelStatusUpdatedNotification) notification)
					.getLevelStatus();
			if (status.hasEnded()) {
				ended.set(true);
				loaded.set(false);
				ships.clear();
				if (status == LevelStatus.COMPLETED) {
					// load next level
					if (levels.size() > levelNumber.getAndIncrement() + 1) {
						synchronized (this) {
							timer.cancel();
							timer = new Timer();
							timer.schedule(new TimerTask() {
								@Override
								public void run() {
									sendRequest(new UpdateLevelMapRequest(
											LevelManagerComponent.this, levels
													.get(levelNumber.get())));
								}
							}, 1000);
						}
					} else {
						levelNumber.set(0);
					}
				} else {
					sendRequest(new UpdateLevelStatusRequest(this,
							LevelStatus.INITIALISING));
					synchronized (this) {
						timer.cancel();
						timer = new Timer();
						timer.schedule(new TimerTask() {
							@Override
							public void run() {
								sendRequest(new UpdateLevelStatusRequest(
										LevelManagerComponent.this,
										LevelStatus.LOADED));
							}
						}, 1000);
					}
				}
			} else if (status == LevelStatus.LOADED) {
				ended.set(false);
				loaded.set(true);
			} else if (status == LevelStatus.STARTED) {
				loaded.set(false);
			}
		}
	}

	@Override
	protected void receiveRequest(C2Request request) {
	}

}
