package uci.inf221.lunarLander.net.connector.clientToServer.component;

import java.io.IOException;
import java.net.InetAddress;

import org.apache.log4j.Logger;

import uci.inf221.lunarLander.arch.Message;
import uci.inf221.lunarLander.arch.c2.C2Component;
import uci.inf221.lunarLander.arch.c2.C2Notification;
import uci.inf221.lunarLander.arch.c2.C2Request;
import uci.inf221.lunarLander.net.connector.clientToServer.message.HandleConnectionStateChangeRequest;
import uci.inf221.lunarLander.net.connector.clientToServer.message.HandleIncomingMessageRequest;
import uci.inf221.lunarLander.net.connector.clientToServer.message.NewMessageNotification;
import uci.inf221.lunarLander.server.Server;
import uci.inf221.lunarLander.utils.KryoInitializer;
import uci.inf221.lunarLander.utils.TerminateableLoopThread;

import com.esotericsoftware.kryonet.Connection;
import com.esotericsoftware.kryonet.Listener;
import com.esotericsoftware.kryonet.Listener.ThreadedListener;

public class ServerConnectionManagerComponent extends C2Component {

	public class ClientListener extends ThreadedListener {

		public ClientListener() {
			super(new Listener() {

				@Override
				public void connected(Connection connection) {
					sendRequest(new HandleConnectionStateChangeRequest(
							ServerConnectionManagerComponent.this, connection
									.getRemoteAddressTCP().getAddress(), true));
					log.info("server connected: " + connection.getID());
				}

				@Override
				public void disconnected(Connection connection) {
					sendRequest(new HandleConnectionStateChangeRequest(
							ServerConnectionManagerComponent.this, null, false));
					log.info("server disconnected: " + connection.getID());

					if (!stop) {
						hostDiscoveryThread = new HostDiscoveryThread();
						hostDiscoveryThread.start();
					}
				}

				@Override
				public void received(Connection connection, Object object) {
					if (object instanceof Message) {
						sendRequest(new HandleIncomingMessageRequest(
								ServerConnectionManagerComponent.this,
								(Message) object));
						log.trace("new message: " + object.toString());
					} else {
						log.warn("discarding unknown message: "
								+ object.toString());
					}
				}

			});
		}

		public void terminate() {
			threadPool.shutdown();
		}

		public boolean isTerminated() {
			return threadPool.isTerminated();
		}

	}

	public static final String PORT_IDENTIFIER_CONNECTION_MANAGER_CONNECTOR = "connection_manager";

	private static final int WRITE_BUFFER_SIZE = 1 << 16;
	private static final int OBJECT_BUFFER_SIZE = 1 << 15;

	private Logger log = Logger.getLogger(this.getClass());

	private com.esotericsoftware.kryonet.Client client;

	private TerminateableLoopThread hostDiscoveryThread;
	private ClientListener clientListener;

	private boolean stop;

	@Override
	protected void createPorts() {
		addTopPort(PORT_IDENTIFIER_CONNECTION_MANAGER_CONNECTOR);
	}

	@Override
	protected void initialize() {
		client = new com.esotericsoftware.kryonet.Client(WRITE_BUFFER_SIZE,
				OBJECT_BUFFER_SIZE);
		KryoInitializer.initKryo(client.getKryo());

		client.addListener((clientListener = new ClientListener()));

		hostDiscoveryThread = new HostDiscoveryThread();
	}

	@Override
	public void start() {
		client.start();
		hostDiscoveryThread.start();
	}

	@Override
	public void stop() {
		stop = true;
		if (hostDiscoveryThread.isAlive())
			hostDiscoveryThread.terminate();
		client.stop();
	}

	@Override
	public void waitForStop() {
		hostDiscoveryThread.waitForTermination();

		clientListener.terminate();
		while (!clientListener.isTerminated()) {
			try {
				Thread.sleep(1);
			} catch (InterruptedException e) {
			}
		}
	}

	@Override
	protected void receiveNotification(C2Notification notification) {
		if (notification instanceof NewMessageNotification) {
			NewMessageNotification newMessageNotification = (NewMessageNotification) notification;
			Message message = newMessageNotification.getMessage();
			if (newMessageNotification.isImportant()) {
				log.trace("sending message to server (TCP): "
						+ message.toString());
				client.sendTCP(message);
			} else {
				log.trace("sending message to server (UDP): "
						+ message.toString());
				client.sendUDP(message);
			}
		}
	}

	@Override
	protected void receiveRequest(C2Request request) {
	}

	private class HostDiscoveryThread extends TerminateableLoopThread {
		public HostDiscoveryThread() {
			super(new Runnable() {
				@Override
				public void run() {
					log.info("sending host discovery packet");
					InetAddress host = client.discoverHost(Server.UDP_PORT,
							1000);
					if (host != null) {
						try {
							client.connect(1000, host, Server.TCP_PORT,
									Server.UDP_PORT);
							hostDiscoveryThread.terminate();
						} catch (IOException e) {
							log.warn(
									"could not connect to server "
											+ host.toString(), e);
						}
					}
				}
			});
		}
	}
}
