package org.cynetec.lib.system.net.ntp;

import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.concurrent.atomic.AtomicBoolean;

import org.apache.commons.net.ntp.NTPUDPClient;
import org.apache.commons.net.ntp.TimeInfo;
import org.apache.log4j.Logger;
import org.cynetec.lib.system.net.ntp.NTPTimeUpdateStatusListener.NTPUpdateStatus;

public class NTPTime extends Thread {
	private Logger log = Logger.getLogger(this.getClass());
	private final SimpleDateFormat simpleDate = new SimpleDateFormat(
			"yyyy-MM-dd HH:mm:ss.SSS");

	private AtomicBoolean terminated;
	private NTPUDPClient ntpClient;
	private long syncInterval;
	private String ntpServer;
	private int ntpServerPort;
	private long relativeTimeBase;
	private long systemTimeBase;
	private NTPTimeUpdateStatusListener listener;

	public NTPTime() {
		this(null, 0L);
	}

	public NTPTime(String ntpServer, long syncInterval) {
		this(ntpServer, NTPUDPClient.DEFAULT_PORT, syncInterval);
	}

	public NTPTime(String ntpServer, int ntpServerPort, long syncInterval) {
		this(ntpServer, NTPUDPClient.DEFAULT_PORT, syncInterval, null);
	}

	public NTPTime(String ntpServer, int ntpServerPort, long syncInterval,
			NTPTimeUpdateStatusListener listener) {
		terminated = new AtomicBoolean(false);
		ntpClient = new NTPUDPClient();
		ntpClient.setDefaultTimeout(5000);
		this.syncInterval = syncInterval;
		this.ntpServer = ntpServer;
		this.ntpServerPort = ntpServerPort;
		this.listener = listener;

		this.setName("NTPTime Thread");
		setSyncTime(System.currentTimeMillis());
	}

	public synchronized void setUpdateStatusListener(
			NTPTimeUpdateStatusListener listener) {
		this.listener = listener;
	}

	public synchronized NTPTimeUpdateStatusListener getUpdateStatusListener() {
		return listener;
	}

	public synchronized void setSyncInterval(long syncInterval) {
		this.syncInterval = syncInterval;
	}

	public synchronized void setNTPServer(String ntpServer) {
		this.ntpServer = ntpServer;
	}

	public synchronized void setNTPServerPort(int ntpServerPort) {
		this.ntpServerPort = ntpServerPort;
	}

	private synchronized void setSyncTime(long relativeTimeBase) {
		this.systemTimeBase = System.currentTimeMillis();
		this.relativeTimeBase = relativeTimeBase;
	}

	@Override
	public void run() {
		while (!terminated.get()) {
			synchronized (this) {
				if (listener != null)
					listener.ntpStatusChange(NTPUpdateStatus.UPDATING);
			}
			
			long syncInterval;
			String ntpServer;
			int ntpServerPort;

			synchronized (this) {
				syncInterval = this.syncInterval;
				ntpServer = this.ntpServer;
				ntpServerPort = this.ntpServerPort;
			}

			try {
				InetAddress host = InetAddress.getByName(ntpServer);

				TimeInfo timeInfo = ntpClient.getTime(host, ntpServerPort);
				timeInfo.computeDetails();

				setSyncTime(System.currentTimeMillis() + timeInfo.getOffset());

				boolean isSNTP = false;
				if (timeInfo.getMessage().getPoll() == 0) {
					isSNTP = true;
				}

				if (log.isInfoEnabled()) {
					log.info("setting reference time to: "
							+ simpleDate.format(new Date(this
									.getNTPTimeMillis()))
							+ " / "
							+ simpleDate.format(new Date())
							+ " from ["
							+ (isSNTP ? "S/NTP" : timeInfo.getMessage()
									.getType()) + " "
							+ timeInfo.getMessage().getVersion() + " "
							+ timeInfo.getMessage().getReferenceIdString()
							+ "]");
				}

				synchronized (this) {
					if (listener != null)
						listener.ntpStatusChange(NTPUpdateStatus.SUCCEEDED);
				}
			} catch (UnknownHostException e) {
				log.error("could not sync time: unknown host " + ntpServer
						+ " - " + e.getMessage());
				synchronized (this) {
					if (listener != null)
						listener.ntpStatusChange(NTPUpdateStatus.FAILED);
				}
			} catch (IOException e) {
				log.error("could not sync time with " + ntpServer + " - "
						+ e.getMessage());
				synchronized (this) {
					if (listener != null)
						listener.ntpStatusChange(NTPUpdateStatus.FAILED);
				}
			}

			try {
				Thread.sleep(syncInterval);
			} catch (InterruptedException e) {
			}
		}
	}

	public void terminate() {
		this.terminated.set(true);
		this.interrupt();
	}

	public synchronized long getNTPTimeMillis() {
		return relativeTimeBase + (System.currentTimeMillis() - systemTimeBase);
	}

}
